/* m68hc05-dis.c -- Motorola 68HC05
   Copyright 2004 Free Software Foundation, Inc.
   Copyright 2004 Martin Kgler
   Contributed by Martin Kgler (mkoegler@auto.tuwien.ac.at)

   This file is part of GDB, GAS, and the GNU binutils.
   
   GDB, GAS, and the GNU binutils are free software; you can redistribute
   them and/or modify them under the terms of the GNU General Public
   License as published by the Free Software Foundation; either version
   2, or (at your option) any later version.
   
   GDB, GAS, and the GNU binutils are distributed in the hope that they
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
   the GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this file; see the file COPYING.  If not, write to the Free
   Software Foundation, 51 Franklin Street - Fifth Floor, Boston, MA
   02110-1301, USA.  */

#include <stdio.h>
#include "sysdep.h"
#include "dis-asm.h"

typedef enum {
  M68HC05_NONE,
  M68HC05_IMM,
  M68HC05_OFFSET8,
  M68HC05_OFFSET16,
  M68HC05_INDEX0,
  M68HC05_INDEX1,
  M68HC05_INDEX2,
  M68HC05_PCREL,
  M68HC05_OFFSET8_PCREL,
} M68HC05_PARAM;

struct m68hc05_opcodes_s
{
  char *name;
  M68HC05_PARAM par;
  unsigned char opcode;
};

#define M68HC05_INS(NAME, PAR, OPCODE) \
{NAME, PAR, OPCODE},

const struct m68hc05_opcodes_s m68hc05_opcodes[] =
{
  #include "opcode/m68hc05.h"
  {NULL, 0, 0}
};


static int
read_memory (bfd_vma memaddr, bfd_byte* buffer, int size,
             struct disassemble_info* info)
{
  int status;

  status = (*info->read_memory_func) (memaddr, buffer, size, info);
  if (status != 0)
    {
      (*info->memory_error_func) (status, memaddr, info);
      return -1;
    }
  return 0;
}


int
print_insn_m68hc05(bfd_vma addr,disassemble_info *info)
{
  int status;
  unsigned char data[5];
  const struct m68hc05_opcodes_s* ops;

  status = read_memory (addr, data, 1, info);
  if (status != 0)
    {
      return status;
    }
  for(ops=m68hc05_opcodes;ops->name;ops++)
    if(ops->opcode==data[0])break;

  if(!ops->name){
    (*info->fprintf_func)(info->stream,".byte %02x",data[0]);
    return 1;
  }
  switch(ops->par){
  case M68HC05_NONE:
    (*info->fprintf_func)(info->stream,"%s",ops->name);
    return 1;
  case M68HC05_IMM:
    status = read_memory (addr+1, data+1, 1, info);
    if (status != 0)
      {
	return status;
      }
    (*info->fprintf_func)(info->stream,"%s\t$0x%x",ops->name,data[1]);
    return 2;
  case M68HC05_OFFSET8:
    status = read_memory (addr+1, data+1, 1, info);
    if (status != 0)
      {
	return status;
      }
    (*info->fprintf_func)(info->stream,"%s\t",ops->name);
    (*info->print_address_func)(data[1],info);
    return 2;
  case M68HC05_OFFSET16:
    status = read_memory (addr+1, data+1, 2, info);
    if (status != 0)
      {
	return status;
      }
    (*info->fprintf_func)(info->stream,"%s\t",ops->name);
    (*info->print_address_func)((data[1]<<8)|data[2],info);
    return 3;
  case M68HC05_INDEX0:
    (*info->fprintf_func)(info->stream,"%s\t%%X",ops->name);
    return 1;
  case M68HC05_INDEX1:
    status = read_memory (addr+1, data+1, 1, info);
    if (status != 0)
      {
	return status;
      }
    (*info->fprintf_func)(info->stream,"%s\t%%X,",ops->name);
    (*info->print_address_func)(data[1],info);
    return 2;
  case M68HC05_INDEX2:
    status = read_memory (addr+1, data+1, 2, info);
    if (status != 0)
      {
	return status;
      }
    (*info->fprintf_func)(info->stream,"%s\t%%X,",ops->name);
    (*info->print_address_func)((data[1]<<8)|data[2],info);
    return 3;
  case M68HC05_PCREL:
    status = read_memory (addr+1, data+1, 1, info);
    if (status != 0)
      {
	return status;
      }
    (*info->fprintf_func)(info->stream,"%s\t",ops->name);
    (*info->print_address_func)(addr+2+(signed char)data[1],info);
    return 2;
  case M68HC05_OFFSET8_PCREL:
    status = read_memory (addr+1, data+1, 2, info);
    if (status != 0)
      {
	return status;
      }
    (*info->fprintf_func)(info->stream,"%s\t",ops->name);
    (*info->print_address_func)(data[1],info);
    (*info->fprintf_func)(info->stream,",");
    (*info->print_address_func)(addr+3+(signed char)data[2],info);
    return 3;
    }
  return 1;
}
